from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .genome import Genome

from .feature import Feature
from .misc_features import Protein
from .utils.genefunctions import reverse_complement

class CDS(Feature):

    __slots__ = (
        'main', 'CDS_segments', 'five_prime_UTR_seq', 'three_prime_UTR_seq',
        'full_UTR_exons', 'protein', 'UTRs'
    )

    protein: Protein|None
    size: int

    def __init__(self, CDS_segments:list, feature_id:str, 
                 ch:str, source:str, feature:str, strand:str, start:int, 
                 end:int, score:str, phase:str, parents:list[str]=[], attributes:dict={}):
        super().__init__(feature_id, ch, source, feature, strand, start, end,
                         score, phase, parents, attributes)    
        self.main = False
        self.CDS_segments = CDS_segments
        #self.kozak = ""
        #self.codon_usage = {}
        self.five_prime_UTR_seq = ""
        self.three_prime_UTR_seq = ""
        self.full_UTR_exons = 0
        self.frame = "."
        self.protein = None
        self.update()

    def update(self):
        self.update_size()
        self.update_phase()
        self.update_frame()

    def update_size(self):
        self.size = 0
        for segment in self.CDS_segments:
            self.size += segment.size

    def update_phase(self):
        leftover = 0
        if self.strand == "+":
            for cs in self.CDS_segments:
                if leftover == 0:
                    cs.phase = 0
                else:
                    cs.phase = 3 - leftover
                leftover = (cs.size - cs.phase) % 3
        elif self.strand == "-":
            for cs in reversed(self.CDS_segments):
                if leftover == 0:
                    cs.phase = 0
                else:
                    cs.phase = 3 - leftover
                leftover = (cs.size - cs.phase) % 3    

    def update_frame(self):
        if self.strand == "+":
            for cs in self.CDS_segments:
                frame = (cs.start + cs.phase) % 3
                if frame == 0:
                    frame = 3
                cs.frame = frame

        if self.strand == "-":
            for cs in reversed(self.CDS_segments):
                frame = (cs.start + cs.phase) % 3
                if frame == 0:
                    frame = 3
                frame = 7 - frame
                cs.frame = frame


    def rename(self, base_id:str, base_gene_id:str, count:int, sep:str="_", digits:int=3, keep_numbering:bool=False, keep_existing_ids_if_derived_from_base_id:bool=False, cds_segment_ids:bool=False):

        rename = False
        rename_cs = False

        if keep_existing_ids_if_derived_from_base_id:
            if base_gene_id not in self.id:
                rename = True
            for cs in self.CDS_segments:
                if base_gene_id not in cs.id:
                    rename_cs = True
        else:
            rename = True
            rename_cs = True

        if rename:
            self.renamed = True

            if keep_numbering and self.id_number != None:
                if self.main:
                    self.id = f"{base_id}{sep}CDS{self.id_number:0{digits}d}"
                else:
                    self.id = f"{base_id}{sep}CDS{self.id_number:0{digits}d}"
            else:
                if self.main:
                    self.id = f"{base_id}{sep}CDS{1:0{digits}d}"
                else:
                    self.id = f"{base_id}{sep}CDS{count:0{digits}d}"

            if self.original_id != self.id:
                self.renamed = True
                self.update_numbering()
                self.generate_protein()

        cs_count = 0
        for cs in self.CDS_segments:
            cs_count += 1

            if rename_cs:

                if keep_numbering and cs.id_number != None:
                    cs.id = f"{base_id}{sep}CDS{cs.id_number}"
                elif cds_segment_ids:
                    cs.id = f"{base_id}{sep}CDS{self.id_number}{sep}{cs_count}"
                else:
                    cs.id = self.id

                if cs.original_id != cs.id:
                    self.renamed = True
                    cs.update_numbering()


    def clear_UTRs(self):
        self.five_prime_UTR_seq = ""
        self.three_prime_UTR_seq = ""
        self.full_UTR_exons = 0
        del self.UTRs

    def generate_sequence(self, genome:Genome, low_memory:bool=False):
        self.seq = ""
        for segment in self.CDS_segments:
            segment.generate_sequence(genome)
        if self.strand == "+":
            for segment in self.CDS_segments:
                self.seq += segment.seq
        elif self.strand == "-":
            for segment in reversed(self.CDS_segments):
                self.seq += segment.seq
        elif self.strand == ".":
            self.seqs = ["", ""]
            for segment in self.CDS_segments:
                self.seqs[0] += segment.seqs[0]
            for segment in reversed(self.CDS_segments):
                self.seqs[1] += segment.seqs[1]

        if low_memory:
            for segment in self.CDS_segments:
                segment.clear_sequence()
        else:
            if hasattr(self, "UTRs"):
                for u in self.UTRs:
                    u.generate_sequence(genome)
                if self.strand == "+":
                    for u in self.UTRs:
                        if u.prime == "5'":
                            self.five_prime_UTR_seq += u.seq
                        else:
                            self.three_prime_UTR_seq += u.seq
                elif self.strand == "-":
                    for u in reversed(self.UTRs):
                        if u.prime == "5'":
                            self.five_prime_UTR_seq += u.seq
                        else:
                            self.three_prime_UTR_seq += u.seq
        self.generate_protein(low_memory=low_memory)

    def generate_hard_sequence(self, hard_masked_genome:Genome, low_memory:bool=False):
        self.hard_seq = ""
        for segment in self.CDS_segments:
            segment.generate_hard_sequence(hard_masked_genome)
        if self.strand == "+":
            for segment in self.CDS_segments:
                self.hard_seq += segment.hard_seq
        elif self.strand == "-":
            for segment in reversed(self.CDS_segments):
                self.hard_seq += segment.hard_seq
        elif self.strand == ".":
            self.hard_seqs = ["", ""]
            for segment in self.CDS_segments:
                self.hard_seqs[0] += segment.hard_seqs[0]
            for segment in reversed(self.CDS_segments):
                self.hard_seqs[1] += segment.hard_seqs[1]

        if low_memory:
            for segment in self.CDS_segments:
                segment.clear_sequence(just_hard=False)
        else:
            if hasattr(self, "UTRs"):
                for u in self.UTRs:
                    u.generate_hard_sequence(hard_masked_genome)

    def clear_sequence(self, just_hard:bool=False, keep_proteins:bool=False):
        self.hard_seq = ""
        for segment in self.CDS_segments:
            segment.clear_sequence(just_hard=just_hard)
        if hasattr(self, "UTRs"):
            for u in self.UTRs:    
                u.clear_sequence(just_hard=just_hard)
        if not just_hard:
            self.seq = ""
            self.five_prime_UTR_seq = ""
            self.three_prime_UTR_seq = ""
            if not keep_proteins:
                self.protein = None

    def generate_protein(self, readthrough:str="both", low_memory:bool=False):
        self.protein = Protein(f"{self.id}.prot", self.seq, self.ch, readthrough)
        if low_memory:
            self.seq = ""

    def equal_segments(self, other:CDS):
        self.CDS_segments.sort()
        other.CDS_segments.sort()
        same = True
        if len(self.CDS_segments) == len(other.CDS_segments):
            for n, segment in enumerate(self.CDS_segments):
                if not segment.equal_sequence(other.CDS_segments[n]):
                    same = False
        else:
            same = False
        
        return same

class Exon(Feature):


    def __init__(self, feature_id:str, ch:str, source:str, feature:str,
                 strand:str, start:int, end:int, score:str, phase:str, parents:list[str]=[], 
                 attributes:dict={}):
        super().__init__(feature_id, ch, source, feature, strand, start, end,
                         score, phase, parents, attributes)

class UTR(Feature):

    __slots__ = ('prime',)
    def __init__(self, feature_id:str, ch:str, source:str, feature:str,
                 strand:str, start:int, end:int, score:str, phase:str, parents:list[str]=[],
                 attributes:dict={}):
        super().__init__(feature_id, ch, source, feature, strand, start, end,
                         score, phase, parents, attributes)
        self.prime = "3'"

class Intron(Feature):
    __slots__ = (
        'intra_coding', 'boundary', 'canonical',
        'splice_site_donor', 'splice_site_acceptor'
    )
    canonical_seqs = ["GT-AG", "GC-AG", "AT-AC"]

    def __init__(self, feature_id:str, ch:str, source:str, feature:str,
                 strand:str, start:int, end:int, score:str, phase:str, parents:list[str]=[],
                 attributes:dict={}):
        super().__init__(feature_id, ch, source, feature, strand, start, end,
                         score, phase, parents, attributes)
        self.intra_coding = False
        self.boundary = ""
        self.canonical = False
        self.splice_site_donor = ""
        self.splice_site_acceptor = ""

    def generate_sequence(self, genome:Genome):
        if self.strand == "+":
            self.seq = genome.scaffolds[self.ch].seq[self.start-1:self.end]
        elif self.strand == "-":
            self.seq = reverse_complement(genome.scaffolds[self.ch].seq[self.start-1:self.end])
        elif self.strand == ".":
            self.seqs = [genome.scaffolds[self.ch].seq[self.start-1:self.end], reverse_complement(genome.scaffolds[self.ch].seq[self.start-1:self.end])]
        self.splice_site_donor = self.seq[0:2]
        self.splice_site_acceptor = self.seq[-2:]
        self.boundary = f"{self.splice_site_donor}-{self.splice_site_acceptor}"
        if self.boundary in Intron.canonical_seqs:
            self.canonical = True
        else:
            self.canonical = False

    def clear_sequence(self, just_hard:bool=False):
        self.hard_seq = ""
        if not just_hard:
            self.seq = ""


from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..annotation import Annotation
    from ..genome import Genome

import time
import os
import warnings
import pandas as pd
import sys

from pathlib import Path
from tqdm import tqdm
from typing import Literal

class AnnotationExport:
    """
    Component for handling export methods for the Annotation class.
    Accessed via 'annotation_object.export'.
    """
    def __init__(self, annotation: Annotation):
        self._annot = annotation

    def all_features(self, feature_output: Literal["main", "all", "both"] = "main", promoters: bool = True, verbose: bool = True, path: str = "", most_specific_id_level = "promoter", quiet: bool = False):
        """
        The "output" parameter can be both, main or all. This parameter only 
        affects promoter, transcript, CDS and protein sequences. If both is selected
        a "main features" file and and "all features" file will be produced.
        """

        feature_output_choices = ["main", "all", "both"]
        if feature_output not in feature_output_choices:
            raise ValueError(f"feature_output={feature_output} is not amongst the feature_output_choices={feature_output_choices} to export all features.")

        most_specific_id_level_choices = ["gene", "transcript", "CDS", "protein", "promoter"]

        if most_specific_id_level not in most_specific_id_level_choices:
            raise ValueError(f"most_specific_id_level={most_specific_id_level} is not amongst the most_specific_id_level_choices={most_specific_id_level_choices} to export proteins.")

        start_time = time.time()

        self.genes(verbose, path)

        if feature_output == "both":
            modes = [True, False]
        elif feature_output == "main":
            modes = [True]
        else:
            modes = [False]

        for b in modes:

            if most_specific_id_level == "promoter" or most_specific_id_level == "protein":
                self.proteins(b, verbose, path)
                self.CDSs(b, verbose, path)
                self.transcripts(b, verbose, path)
                if promoters:
                    self.promoters(b, verbose, path)

            elif most_specific_id_level == "CDS":
                self.proteins(b, verbose, path, used_id="CDS")
                self.CDSs(b, verbose, path)
                self.transcripts(b, verbose, path)
                if promoters:
                    self.promoters(b, verbose, path)

            elif most_specific_id_level == "transcript":
                self.proteins(b, verbose, path, used_id="transcript")
                self.CDSs(b, verbose, path, used_id="transcript")
                self.transcripts(b, verbose, path)
                if promoters:
                    self.promoters(b, verbose, path, used_id="transcript")

            elif most_specific_id_level == "gene":
                self.proteins(b, verbose, path, used_id="gene")
                self.CDSs(b, verbose, path, used_id="gene")
                self.transcripts(b, verbose, path, used_id="gene")
                if promoters:
                    self.promoters(b, verbose, path, used_id="gene")

        now = time.time()
        lapse = now - start_time
        if not quiet:
            print(f"Extracting {self._annot.id} annotation features took {round(lapse, 1)} seconds\n")

    def proteins(self, only_main: bool = True, verbose: bool = True, custom_path: str = "", used_id: str = "protein", unique_proteins_per_gene: bool = False, only_cds_main: bool = True):
        """
        Main proteins means only proteins obtained from the main CDSs of the
        main transcripts. This equates to one protein per gene.

        Verbose will include protein tag details and readthrough options, strand, chromosome, and coordinates.

        Proteins will be exported into fastas with their protein ids unless 
        CDS, transcript or gene has been selected as used_id. To avoid duplicate
        equal entries, when choosing gene or transcript only main proteins will
        be able to be output.

        valid_id_choices = ["gene", "transcript", "CDS", "protein"]
        """

        if custom_path:
            output_file = Path(custom_path)
        else:
            output_file = Path(self._annot.path) / "features"
        output_file.mkdir(parents=True, exist_ok=True)
        output_file = str(output_file) + "/"

        output_file += self._annot.id
        output_file += self._annot.feature_suffix
        output_file += "_proteins"

        out = ""

        if unique_proteins_per_gene:
            only_main = False
            only_cds_main = False
            if used_id == "gene":
                used_id = "protein"
                warnings.warn(f"Used id 'gene' has been changed to 'protein' as unique proteins per gene was selected.", category=UserWarning)

        if used_id == "gene":
            only_main = True
            only_cds_main = True
            output_file += "_g_id_main"
        else:
            if used_id == "transcript":
                only_cds_main = True
                output_file += "_t_id"
                if unique_proteins_per_gene:
                    warnings.warn(f"If more than one CDS exists per transcript (this is rarely the case), CDSs beyond the main CDS will not be considered, since 'transcript' was the used_id. Select 'CDS' or 'protein' if all CDSs are to be considered.", category=UserWarning)
            elif used_id == "CDS":
                output_file += "_c_id"
            elif used_id == "protein":
                output_file += "_p_id"

            if unique_proteins_per_gene:
                output_file += "_unique_per_gene"
            elif only_main:
                output_file += "_main"
            else:
                output_file += "_all"

        if verbose:
            output_file += "_coordinates"
            
        output_file += ".fasta"
        
        valid_id_choices = ["gene", "transcript", "CDS", "protein"]
        if used_id not in valid_id_choices:
            raise ValueError(f"used_id={used_id} is not amongst the valid_id_choices={valid_id_choices} to export proteins.")
        
        for genes in self._annot.chrs.values():
            for g in genes.values():
                temp_cs = []

                for t in g.transcripts.values():
                    if only_main:
                        if t.main:
                            for c in t.CDSs.values():
                                if c.seq != "":
                                    if only_cds_main:
                                        if c.main:
                                            temp_cs.append(c)
                                    else:
                                        temp_cs.append(c)
                    else:
                        for c in t.CDSs.values():
                            if c.seq != "":
                                if only_cds_main:
                                    if c.main:
                                        temp_cs.append(c)
                                else:
                                    temp_cs.append(c)

                if not unique_proteins_per_gene:
                    final_cs = temp_cs.copy()

                else:
                    final_cs = []
                    if len(temp_cs) > 0:
                        final_cs.append(temp_cs[0])

                    for i, c1 in enumerate(temp_cs):
                        if i > 0:
                            add = True
                            for c2 in final_cs:
                                if c1.equal_segments(c2):
                                    add = False
                            if add:
                                final_cs.append(c1)

                for c in final_cs:

                    if used_id == "protein":
                        out += f">{c.protein.id}"
                    elif used_id == "CDS":
                        out += f">{c.id}"
                    elif used_id == "transcript":
                        out += f">{c.parents[0]}"
                    elif used_id == "gene":
                        out += f">{g.id}"

                    if c.protein.summary_tag and verbose:
                        out += f"|{c.protein.summary_tag}"
                    if verbose:
                        out += f"|readthrough:{c.protein.readthrough}|{c.strand}|{c.ch}|{c.start}:{c.end}"

                    out += f"\n{c.protein.seq}\n"        
                     
        if out != "":
            f_out = open(output_file, "w", encoding="utf-8")
            f_out.write(out)
            f_out.close()
        else:
            print(f"Warning: Run self.generate_sequences(genome) on {self._annot.id}")

    def unique_proteins(self, genome: Genome | None = None, custom_path: str = "", quiet: bool = False):
        start_time = time.time()
        # Check if stdout or stderr are redirected to files
        stdout_redirected = not sys.stdout.isatty()
        stderr_redirected = not sys.stderr.isatty()

        # Disable tqdm if stdout or stderr are redirected
        if stdout_redirected or stderr_redirected or quiet:
            disable = True
        else:
            disable = False
        
        if not self._annot.contains_protein_sequences:
            if genome == None:
                print(f"You forgot to provide the genome for {self._annot.id} and no sequences exist for CDSs")
            else:
                self._annot.generate_sequences(genome, quiet=quiet)
                self.unique_proteins(custom_path=custom_path)
        else:
            if custom_path:
                output_file = Path(custom_path)
            else:
                output_file = Path(self._annot.path) / "features"
            output_file.mkdir(parents=True, exist_ok=True)
            output_file = str(output_file) + "/"
            output_file += f"{self._annot.id}{self._annot.feature_suffix}_unique_proteins.fasta"
            out = ""
            all_protein_seqs = {}
            self._annot.all_protein_ids = {}
            for chrom, genes in self._annot.chrs.items():
                for g in genes.values():
                    if g.coding:
                        for t in g.transcripts.values():
                            for c in t.CDSs.values():
                                if c.protein is not None:
                                    if c.protein.seq != "":
                                        all_protein_seqs[c.protein.id] = c.protein.seq
                                        self._annot.all_protein_ids[c.protein.id] = (chrom, g.id, t.id, c.id)

            progress_bar = tqdm(total=len(all_protein_seqs.keys()), disable=disable,
                            bar_format=(
                f'\033[1;91mDetermining and exporting unique {self._annot.id} proteins:\033[0m '
                '{percentage:3.0f}%|'
                f'\033[1;91m{{bar}}\033[0m| '
                '{n}/{total} [{elapsed}<{remaining}]'))
            
            
            unique_sequences = {}
            self._annot.protein_equivalences = {}

            for protein_id, sequence in all_protein_seqs.items():
                progress_bar.update(1)
                if sequence not in unique_sequences:
                    unique_sequences[sequence] = protein_id
                    self._annot.protein_equivalences[protein_id] = []
                else:
                    first_protein_id = unique_sequences[sequence]
                    self._annot.protein_equivalences[first_protein_id].append(protein_id)

            for sequence, protein_id in unique_sequences.items():
                out += f">{protein_id}\n{sequence}\n"

            f_out = open(output_file, "w", encoding="utf-8")
            f_out.write(out)
            f_out.close()

            progress_bar.close()

            now = time.time()
            lapse = now - start_time
            if not quiet:
                print(f"\nExporting unique {self._annot.id} proteins took {round(lapse/60, 1)} minutes")

    def unique_transcripts(self, genome: Genome | None = None, custom_path: str = "", quiet: bool = False, rna_classes: list = []):
        start_time = time.time()
        # Check if stdout or stderr are redirected to files
        stdout_redirected = not sys.stdout.isatty()
        stderr_redirected = not sys.stderr.isatty()

        # Disable tqdm if stdout or stderr are redirected
        if stdout_redirected or stderr_redirected or quiet:
            disable = True
        else:
            disable = False
        
        if not self._annot.contains_all_sequences:
            if genome == None:
                print(f"You forgot to provide the genome for {self._annot.id} and no sequences exist for transcripts")
            else:
                self._annot.generate_sequences(genome, quiet=quiet)
                self.unique_transcripts(custom_path=custom_path, rna_classes=rna_classes)
        else:
            if custom_path:
                output_file = Path(custom_path)
            else:
                output_file = Path(self._annot.path) / "features"
            output_file.mkdir(parents=True, exist_ok=True)
            output_file = str(output_file) + "/"
            output_file += f"{self._annot.id}{self._annot.feature_suffix}_unique_transcripts.fasta"
            out = ""
            all_transcript_seqs = {}
            for chrom, genes in self._annot.chrs.items():
                for g in genes.values():
                    for t in g.transcripts.values():
                        if (t.feature in rna_classes) or (not rna_classes):
                            if t.seq != "":
                                all_transcript_seqs[t.id] = t.seq

            progress_bar = tqdm(total=len(all_transcript_seqs.keys()), disable=disable,
                            bar_format=(
                f'\033[1;91mDetermining and exporting unique {self._annot.id} transcripts:\033[0m '
                '{percentage:3.0f}%|'
                f'\033[1;91m{{bar}}\033[0m| '
                '{n}/{total} [{elapsed}<{remaining}]'))
            
            
            unique_sequences = {}

            for transcript_id, sequence in all_transcript_seqs.items():
                progress_bar.update(1)
                if sequence not in unique_sequences:
                    unique_sequences[sequence] = transcript_id

            for sequence, transcript_id in unique_sequences.items():
                out += f">{transcript_id}\n{sequence}\n"

            f_out = open(output_file, "w", encoding="utf-8")
            f_out.write(out)
            f_out.close()

            progress_bar.close()

            now = time.time()
            lapse = now - start_time
            if not quiet:
                print(f"\nExporting unique {self._annot.id} transcripts took {round(lapse/60, 1)} minutes")

    def unique_CDSs(self, genome: Genome | None = None, custom_path: str = "", quiet: bool = False):
        start_time = time.time()
        # Check if stdout or stderr are redirected to files
        stdout_redirected = not sys.stdout.isatty()
        stderr_redirected = not sys.stderr.isatty()

        # Disable tqdm if stdout or stderr are redirected
        if stdout_redirected or stderr_redirected or quiet:
            disable = True
        else:
            disable = False
        
        if not self._annot.contains_CDS_sequences:
            if genome == None:
                print(f"You forgot to provide the genome for {self._annot.id} and no sequences exist for CDSs")
            else:
                self._annot.generate_sequences(genome, quiet=quiet)
                self.unique_CDSs(custom_path=custom_path)
        else:
            if custom_path:
                output_file = Path(custom_path)
            else:
                output_file = Path(self._annot.path) / "features"
            output_file.mkdir(parents=True, exist_ok=True)
            output_file = str(output_file) + "/"
            output_file += f"{self._annot.id}{self._annot.feature_suffix}_unique_CDSs.fasta"
            out = ""
            all_CDS_seqs = {}
            for chrom, genes in self._annot.chrs.items():
                for g in genes.values():
                    if g.coding:
                        for t in g.transcripts.values():
                            for c in t.CDSs.values():
                                if c.seq != "":
                                    all_CDS_seqs[c.id] = c.seq

            progress_bar = tqdm(total=len(all_CDS_seqs.keys()), disable=disable,
                            bar_format=(
                f'\033[1;91mDetermining and exporting unique {self._annot.id} CDSs:\033[0m '
                '{percentage:3.0f}%|'
                f'\033[1;91m{{bar}}\033[0m| '
                '{n}/{total} [{elapsed}<{remaining}]'))
            
            
            unique_sequences = {}

            for CDS_id, sequence in all_CDS_seqs.items():
                progress_bar.update(1)
                if sequence not in unique_sequences:
                    unique_sequences[sequence] = CDS_id

            for sequence, CDS_id in unique_sequences.items():
                out += f">{CDS_id}\n{sequence}\n"

            f_out = open(output_file, "w", encoding="utf-8")
            f_out.write(out)
            f_out.close()

            progress_bar.close()

            now = time.time()
            lapse = now - start_time
            if not quiet:
                print(f"\nExporting unique {self._annot.id} CDSs took {round(lapse/60, 1)} minutes")

    def CDSs(self, only_main: bool = True, verbose: bool = True, custom_path: str = "", used_id: str = "CDS", unique_CDSs_per_gene: bool = False, only_cds_main: bool = True):
        """
        Main CDSs means only CDS sequence obtained from the main CDS of the
        main transcripts.

        Verbose will include strand, chromosome, and coordinates.

        CDSs will be exported into fastas with their CDS ids unless 
        transcript or gene has been selected as used_id. To avoid duplicate
        equal entries, when choosing gene or transcript only main CDSs will
        be able to be output.

        valid_id_choices = ["gene", "transcript", "CDS"]
        """

        if custom_path:
            output_file = Path(custom_path)
        else:
            output_file = Path(self._annot.path) / "features"
        output_file.mkdir(parents=True, exist_ok=True)
        output_file = str(output_file) + "/"

        output_file += self._annot.id
        output_file += self._annot.feature_suffix
        output_file += "_CDSs"

        out = ""

        if unique_CDSs_per_gene:
            only_main = False
            only_cds_main = False
            if used_id == "gene":
                used_id = "CDS"
                warnings.warn(f"Used id 'gene' has been changed to 'CDS' as unique CDSs per gene was selected.", category=UserWarning)

        if used_id == "gene":
            only_main = True
            only_cds_main = True
            output_file += "_g_id_main"
        else:
            if used_id == "transcript":
                only_cds_main = True
                output_file += "_t_id"
                if unique_CDSs_per_gene:
                    warnings.warn(f"If more than one CDS exists per transcript (this is rarely the case), CDSs beyond the main CDS will not be considered, since 'transcript' was the used_id. Select 'CDS' if all CDSs are to be considered.", category=UserWarning)
            elif used_id == "CDS":
                output_file += "_c_id"

            if unique_CDSs_per_gene:
                output_file += "_unique_per_gene"
            elif only_main:
                output_file += "_main"
            else:
                output_file += "_all"

        if verbose:
            output_file += "_coordinates"
            
        output_file += ".fasta"
        
        valid_id_choices = ["gene", "transcript", "CDS"]
        if used_id not in valid_id_choices:
            raise ValueError(f"used_id={used_id} is not amongst the valid_id_choices={valid_id_choices} to export CDSs.")

        for genes in self._annot.chrs.values():
            for g in genes.values():
                temp_cs = []
                for t in g.transcripts.values():
                    if only_main:
                        if t.main:
                            for c in t.CDSs.values():
                                if c.seq != "":
                                    if only_cds_main:
                                        if c.main:
                                            temp_cs.append(c)
                                    else:
                                        temp_cs.append(c)
                    else:
                        for c in t.CDSs.values():
                            if c.seq != "":
                                if only_cds_main:
                                    if c.main:
                                        temp_cs.append(c)
                                else:
                                    temp_cs.append(c)

                if not unique_CDSs_per_gene:
                    final_cs = temp_cs.copy()
                else:
                    final_cs = []

                    if len(temp_cs) > 0:
                        final_cs.append(temp_cs[0])

                    for i, c1 in enumerate(temp_cs):
                        if i > 0:
                            add = True
                            for c2 in final_cs:
                                if c1.equal_segments(c2):
                                    add = False
                            if add:
                                final_cs.append(c1)

                for c in final_cs:

                    if used_id == "CDS":
                        out += f">{c.id}"
                    elif used_id == "transcript":
                        out += f">{c.parents[0]}"
                    elif used_id == "gene":
                        out += f">{g.id}"

                    if verbose:
                        out += f"|{c.strand}|{c.ch}|{c.start}:{c.end}"

                    out += f"\n{c.seq}\n"
                     
        if out != "":
            f_out = open(output_file, "w", encoding="utf-8")
            f_out.write(out)
            f_out.close()
        else:
            print(f"Warning: Run self.generate_sequences(genome) on {self._annot.id}")

    def transcripts(self, only_main: bool = True, verbose: bool = True, custom_path: str = "", used_id: str = "transcript", rna_classes: list = [], unique_transcripts_per_gene: bool = False):
        """
        Main means only main transcript sequences are exported.

        Verbose will include strand, chromosome, and coordinates.

        Transcripts will be exported into fastas with their transcript ids unless 
        gene has been selected as used_id. To avoid duplicate
        equal entries, when choosing gene only main transcripts will
        be able to be output.

        valid_id_choices = ["gene", "transcript"]
        """
        if custom_path:
            output_file = Path(custom_path)
        else:
            output_file = Path(self._annot.path) / "features"
        output_file.mkdir(parents=True, exist_ok=True)
        output_file = str(output_file) + "/"
        output_file += self._annot.id
        output_file += self._annot.feature_suffix
        output_file += "_transcripts"

        out = ""

        if unique_transcripts_per_gene:
            only_main = False
            if used_id == "gene":
                used_id = "transcript"
                warnings.warn(f"Used id 'gene' has been changed to 'transcript' as unique transcripts per gene was selected.", category=UserWarning)

        if used_id == "gene":
            only_main = True
            output_file += "_g_id_main"
        elif used_id == "transcript":
            output_file += "_t_id"
            if unique_transcripts_per_gene:
                output_file += "_unique_per_gene"
            elif only_main:
                output_file += "_main"
            else:
                output_file += "_all"

        if verbose:
            output_file += "_coordinates"
            
        output_file += ".fasta"

        valid_id_choices = ["gene", "transcript"]
        if used_id not in valid_id_choices:
            raise ValueError(f"used_id={used_id} is not amongst the valid_id_choices={valid_id_choices} to export transcripts.")
        
        for genes in self._annot.chrs.values():
            for g in genes.values():
                temp_ts = []
                for t in g.transcripts.values():
                    if (t.feature in rna_classes) or (not rna_classes):
                        if t.seq != "":
                            if only_main:
                                if t.main:
                                    temp_ts.append(t)
                            else:
                                temp_ts.append(t)

                if not unique_transcripts_per_gene:
                    final_ts = temp_ts.copy()
                else:
                    final_ts = []
                    if len(temp_ts) > 0:
                        final_ts.append(temp_ts[0])

                    for i, t1 in enumerate(temp_ts):
                        if i > 0:
                            add = True
                            for t2 in final_ts:
                                if t1.seq == t2.seq:
                                    add = False
                            if add:
                                final_ts.append(t1)

                for t in final_ts:
                    if used_id == "transcript":    
                        out += f">{t.id}"
                    elif used_id == "gene":
                        out += f">{g.id}"

                    if verbose:
                        out += f"|{t.strand}|{t.ch}|{t.start}:{t.end}"
                    out += f"\n{t.seq}\n"

        if out != "":
            f_out = open(output_file, "w", encoding="utf-8")
            f_out.write(out)
            f_out.close()
        else:
            print(f"Warning: Run self.generate_sequences(genome) on {self._annot.id}")

    def genes(self, verbose: bool = True, custom_path: str = ""):
        if custom_path:
            output_file = Path(custom_path)
        else:
            output_file = Path(self._annot.path) / "features"
        output_file.mkdir(parents=True, exist_ok=True)
        output_file = str(output_file) + "/"
        output_file += self._annot.id
        output_file += self._annot.feature_suffix
        output_file += "_genes"


        if verbose:
            output_file += "_coordinates"
        output_file += ".fasta"
        out = ""

        for genes in self._annot.chrs.values():
            for g in genes.values():
                if g.seq != "":
                    out += f">{g.id}"
                    if verbose:
                        out += f"|{g.strand}|{g.ch}|{g.start}:{g.end}"
                    out += f"\n{g.seq}\n"
        if out != "":
            f_out = open(output_file, "w", encoding="utf-8")
            f_out.write(out)
            f_out.close()
        else:
            print(f"Warning: Run self.generate_sequences(genome) on {self._annot.id}")

    def promoters(self, only_main: bool = True, verbose: bool = True, custom_path: str = "", used_id: str = "promoter"):
        """

        Verbose will include promoter type, strand, chromosome, and coordinates.

        """
        if custom_path:
            output_file = Path(custom_path)
        else:
            output_file = Path(self._annot.path) / "features"
        output_file.mkdir(parents=True, exist_ok=True)
        output_file = str(output_file) + "/"
        output_file += self._annot.id
        output_file += self._annot.feature_suffix
        output_file += "_promoters"

        out = ""

        if used_id == "gene":
            only_main = True
            output_file += "_g_id_main"
        else:
            if used_id == "transcript":
                output_file += "_t_id"
            if used_id == "promoter":
                output_file += "_prom_id"
            if only_main:
                output_file += "_main"
            else:
                output_file += "_all"

        if verbose:
            output_file += "_coordinates"
            
        output_file += f"_{self._annot.promoter_size}_{self._annot.promoter_types}.fasta"

        valid_id_choices = ["gene", "transcript", "promoter"]

        if used_id not in valid_id_choices:
            raise ValueError(f"used_id={used_id} is not amongst the valid_id_choices={valid_id_choices} to export promoters.")

        for genes in self._annot.chrs.values():
            for g in genes.values():
                for t in g.transcripts.values():
                    if only_main and not t.main:
                        continue

                    if hasattr(t, "promoter"):
                        if t.promoter.seq != "":

                            if used_id == "promoter":
                                out += f">{t.promoter.id}"
                            elif used_id == "transcript":
                                out += f">{t.id}"
                            elif used_id == "gene":
                                out += f">{g.id}"

                            if verbose:
                                out += f"|{t.promoter.type}|{t.strand}|{t.ch}|{t.promoter.start}:{t.promoter.end}"

                            out += f"\n{t.promoter.seq}\n"
                        else:
                            print(f"Warning: transcript {t.id} from annotation {self._annot.id} has no promoter sequence.\n")
                    else:
                        print(f"Warning: transcript {t.id} from annotation {self._annot.id} has no promoter.\n")

        if out != "":
            f_out = open(output_file, "w", encoding="utf-8")
            f_out.write(out)
            f_out.close()
        else:
            print(f"Warning: Run self.generate_promoters(genome) and self.generate_sequences(genome) on {self._annot.id}")

    def for_dapseq(self, genome: Genome, genome_out_folder: str = "", gff_out_folder: str = "", tag: str = "_for_dap.gff3", skip_atypical_fts: bool = True, main_only: bool = False, UTRs: bool = False, exclude_non_coding: bool = False):
        equivalences = genome.rename_features_dap(output_folder=genome_out_folder, return_equivalences=True, export=True)
        self._annot.rename_chromosomes(equivalences)
        self.gff(custom_path=gff_out_folder, tag=tag, skip_atypical_fts=skip_atypical_fts, main_only=main_only, UTRs=UTRs, just_genes=exclude_non_coding)

    def gff(self, custom_path: str = "", tag: str = ".gff3", skip_atypical_fts: bool = False, main_only: bool = False, UTRs: bool = False, just_genes: bool = False, no_1bp_features: bool = False, repeat_exons_utrs: bool = False, subfolder: bool = True, quiet: bool = False, skip_orphaned_fts: bool = False, featurecountsID: bool = False, extra_attributes:bool = False, clean_attributes:bool=True, aliases:bool=False, symbols:bool=False, symbols_as_description:bool=False):

        # Check if stdout or stderr are redirected to files
        stdout_redirected = not sys.stdout.isatty()
        stderr_redirected = not sys.stderr.isatty()

        # Disable tqdm if stdout or stderr are redirected
        if stdout_redirected or stderr_redirected or quiet:
            disable = True
        else:
            disable = False

        progress_bar = tqdm(total=len(self._annot.all_gene_ids.keys()), disable=disable,
                                bar_format=(
                    f'\033[38;2;46;204;113m\033[1mExporting gff for {self._annot.id} annotation:\033[0m '
                    '{percentage:3.0f}%|'
                    f'\033[38;2;46;204;113m{{bar}}\033[0m| '
                    '{n}/{total} [{elapsed}<{remaining}]'))

        if subfolder:
            export_folder = Path(custom_path or self._annot.path) / "out_gffs"
        else:
            export_folder = Path(custom_path or self._annot.path)
        export_folder.mkdir(parents=True, exist_ok=True)
        export_folder = str(export_folder) + "/"

        output_suffix = ""

        if just_genes:
            output_suffix += "_just_genes"
        if no_1bp_features:
            output_suffix += "_for_lifton"

        if tag == ".gff3":
            tag = f"{self._annot.id}{self._annot.suffix}{output_suffix}{tag}"
            if not quiet:
                print(f"Exporting {self._annot.id} gff with tag='{tag}' which is dapfit={self._annot.dapfit} and dapmod={self._annot.dapmod} and combined={self._annot.combined}.")
        elif not quiet:
            print(f"Exporting {self._annot.id} gff to {export_folder}{tag}.")

        with open(f"{export_folder}{tag}", "w", encoding="utf-8") as f_out:
            if self._annot.clean or self._annot.dapmod:
                f_out.write("##gff-version 3\n")
            else:
                f_out.write("\n".join(self._annot.gff_header) + "\n")

            if repeat_exons_utrs:
                self._annot.single_parent_for_exons_utrs()

            if featurecountsID:
                self._annot.create_featurecounts_ids()

            for x1, genes in enumerate(self._annot.chrs.values()):
                progress_bar.update(len(genes))
                for x2, g in enumerate(genes.values()):
                    
                    if no_1bp_features:
                        gene_1bp_feature = False
                        for t in g.transcripts.values():
                            for e in t.exons:
                                if e.size == 1:
                                    gene_1bp_feature = True
                            for c in t.CDSs.values():
                                for cs in c.CDS_segments:
                                    if cs.size == 1:
                                        gene_1bp_feature = True
                                for u in c.UTRs:
                                    if u.size == 1:
                                        gene_1bp_feature = True

                        if gene_1bp_feature:
                            continue

                    f_out.write(g.print_gff(extra_attributes=extra_attributes, clean=clean_attributes, aliases=aliases, symbols=symbols, symbols_as_description=symbols_as_description))

                    if just_genes:
                        continue

                    if repeat_exons_utrs or main_only or len(g.transcripts) == 1:
                        for t in g.transcripts.values():
                            if main_only:
                                if not t.main:
                                    continue
                            f_out.write(t.print_gff(featurecountsID=featurecountsID, clean=clean_attributes, aliases=aliases, symbols=symbols, symbols_as_description=symbols_as_description))
                            for e in t.exons:
                                f_out.write(e.print_gff(featurecountsID=featurecountsID, clean=clean_attributes, aliases=aliases, symbols=symbols, symbols_as_description=symbols_as_description))

                            for m in t.miRNAs:
                                f_out.write(m.print_gff(featurecountsID=featurecountsID, clean=clean_attributes, aliases=aliases, symbols=symbols, symbols_as_description=symbols_as_description))

                            for c in t.CDSs.values():
                                if main_only:
                                    if not c.main:
                                        continue
                                for c_seg in c.CDS_segments:
                                    f_out.write(c_seg.print_gff(featurecountsID=featurecountsID, clean=clean_attributes, aliases=aliases, symbols=symbols, symbols_as_description=symbols_as_description))

                                if UTRs:
                                    if hasattr(c, "UTRs"):
                                        for u in c.UTRs:
                                            f_out.write(u.print_gff(featurecountsID=featurecountsID, clean=clean_attributes, aliases=aliases, symbols=symbols, symbols_as_description=symbols_as_description))

                    else:
                                
                        for t in g.transcripts.values():                                
                            f_out.write(t.print_gff(featurecountsID=featurecountsID, clean=clean_attributes, aliases=aliases, symbols=symbols, symbols_as_description=symbols_as_description))

                        exons = []
                        for t in g.transcripts.values():
                            exons.extend(t.exons)

                        exons.sort()
                        
                        unique_exons = []
                        if exons:
                            unique_exons.append(exons[0])
                            for i in range(1, len(exons)):
                                if not exons[i].equal_sequence(exons[i-1]):
                                    unique_exons.append(exons[i])

                        for e in unique_exons:
                            f_out.write(e.print_gff(featurecountsID=featurecountsID, clean=clean_attributes, aliases=aliases, symbols=symbols, symbols_as_description=symbols_as_description))

                        for m in t.miRNAs:
                            f_out.write(m.print_gff(featurecountsID=featurecountsID, clean=clean_attributes, aliases=aliases, symbols=symbols, symbols_as_description=symbols_as_description))

                        for t in g.transcripts.values():
                            for c in t.CDSs.values():
                                for c_seg in c.CDS_segments:
                                    f_out.write(c_seg.print_gff(featurecountsID=featurecountsID, clean=clean_attributes, aliases=aliases, symbols=symbols, symbols_as_description=symbols_as_description))
                        if UTRs:
                            utrs = []
                            for t in g.transcripts.values():
                                for c in t.CDSs.values():
                                    if hasattr(c, "UTRs"):
                                        utrs.extend(c.UTRs)
                            utrs.sort()

                            unique_utrs = []
                            if utrs:
                                unique_utrs.append(utrs[0])
                                for i in range(1, len(utrs)):
                                    if not utrs[i].equal_sequence(utrs[i-1]):
                                        unique_utrs.append(utrs[i])

                            for u in unique_utrs:
                                f_out.write(u.print_gff(featurecountsID=featurecountsID))

                    if x1 == (len(self._annot.chrs) - 1) and x2 == (len(genes) - 1):
                        continue
                    f_out.write("###\n")

            progress_bar.close()

            if not skip_atypical_fts:
                if not just_genes:
                    for x, ft in enumerate(self._annot.atypical_features):
                        if x == 0:
                            f_out.write("###\n")
                        f_out.write(ft.print_gff())
                        if x == (len(self._annot.atypical_features) - 1):
                            continue
                        f_out.write("###\n")
            if not skip_orphaned_fts:
                if not just_genes:
                    for x, ft in enumerate(self._annot.orphaned_features):
                        if x == 0:
                            f_out.write("###\n")
                        f_out.write(ft.print_gff())
                        if x == (len(self._annot.orphaned_features) - 1):
                            continue
                        f_out.write("###\n")

    def gtf(self, custom_path: str = "", tag: str = ".gtf", main_only: bool = False, UTRs: bool = False, just_genes: bool = False, no_1bp_features: bool = False, quiet: bool = False):

        self._annot.create_gtf_attributes()

        # Check if stdout or stderr are redirected to files
        stdout_redirected = not sys.stdout.isatty()
        stderr_redirected = not sys.stderr.isatty()

        # Disable tqdm if stdout or stderr are redirected
        if stdout_redirected or stderr_redirected or quiet:
            disable = True
        else:
            disable = False

        progress_bar = tqdm(total=len(self._annot.all_gene_ids.keys()), disable=disable,
                                bar_format=(
                    f'\033[38;2;46;204;113m\033[1mExporting gtf for {self._annot.id} annotation:\033[0m '
                    '{percentage:3.0f}%|'
                    f'\033[38;2;46;204;113m{{bar}}\033[0m| '
                    '{n}/{total} [{elapsed}<{remaining}]'))

        export_folder = Path(custom_path or self._annot.path) / "out_gtfs"
        export_folder.mkdir(parents=True, exist_ok=True)
        export_folder = str(export_folder) + "/"

        output_suffix = ""

        if just_genes:
            output_suffix += "_just_genes"
        if no_1bp_features:
            output_suffix += "_for_lifton"

        if tag == ".gtf":
            tag = f"{self._annot.id}{self._annot.suffix}{output_suffix}{tag}"
            if not quiet:
                print(f"Exporting {self._annot.id} gtf with tag='{tag}' which is dapfit={self._annot.dapfit} and dapmod={self._annot.dapmod} and combined={self._annot.combined}.")
        elif not quiet:
            print(f"Exporting {self._annot.id} gtf to {export_folder}{tag}.")

        with open(f"{export_folder}{tag}", "w", encoding="utf-8") as f_out:
            f_out.write("#gtf-version 2.2\n")

            for x1, genes in enumerate(self._annot.chrs.values()):
                progress_bar.update(len(genes))
                for x2, g in enumerate(genes.values()):

                    if no_1bp_features:
                        gene_1bp_feature = False
                        for t in g.transcripts.values():
                            for e in t.exons:
                                if e.size == 1:
                                    gene_1bp_feature = True
                            for c in t.CDSs.values():
                                for cs in c.CDS_segments:
                                    if cs.size == 1:
                                        gene_1bp_feature = True
                                for u in c.UTRs:
                                    if u.size == 1:
                                        gene_1bp_feature = True

                        if gene_1bp_feature:
                            continue

                    f_out.write(g.print_gtf())

                    if just_genes:
                        continue

                    for t in g.transcripts.values():
                        if main_only:
                            if not t.main:
                                continue
                        original_feature = t.feature
                        t.feature = "transcript"
                        f_out.write(t.print_gtf())
                        t.feature = original_feature
                        for e in t.exons:
                            f_out.write(e.print_gtf())
                        for c in t.CDSs.values():
                            if main_only:
                                if not c.main:
                                    continue
                            for c_seg in c.CDS_segments:
                                f_out.write(c_seg.print_gtf())
                            if UTRs:
                                if hasattr(c, "UTRs"):
                                    for u in c.UTRs:
                                        f_out.write(u.print_gtf())

                    if x1 == (len(self._annot.chrs) - 1) and x2 == (len(genes) - 1):
                        continue
                    f_out.write("###\n")

            progress_bar.close()

    def gene_list(self, custom_path: str = "", output_file: str = "", lengths: bool = False, coordinates: bool = False, chromosomes: bool = False, coding_info: bool = False, skip_coding: bool = False, skip_non_coding: bool = False, sep: str = "\t", skip_pseudogenes: bool = False, skip_transposables: bool = False, gene_symbols: bool = False):

        if not custom_path:
            export_folder = Path(self._annot.path) / "lists"
        else:
            export_folder = Path(custom_path)
        export_folder.mkdir(parents=True, exist_ok=True)
        export_folder = str(export_folder) + "/"

        if not output_file:
            output_file = f"{export_folder}{self._annot.name}_genes.txt"
        else:
            output_file = f"{export_folder}{output_file}"

        if skip_coding and skip_non_coding:
            print("Warning: Both skip_coding and skip_non_coding are set to True. No genes will be listed.")
            return

        with open(output_file, "w", encoding="utf-8") as f_out:

            header = ["gene_id"]
            if chromosomes or coordinates:
                header.append("scaffold")
            if coordinates:
                header.append("gene_start")
                header.append("gene_end")
            if lengths:
                header.append("gene_length")
            if coding_info:
                header.append("coding")
            if gene_symbols:
                header.append("gene_symbol")
            f_out.write(sep.join(header) + "\n")

            for chrom, genes in self._annot.chrs.items():
                for g in genes.values():
                    if skip_pseudogenes and g.pseudogene:
                        continue
                    if skip_transposables and g.transposable:
                        continue
                    if skip_coding and g.coding:
                        continue
                    if skip_non_coding and not g.coding:
                        continue

                    out = [g.id]
                    if chromosomes or coordinates:
                        out.append(chrom)
                    if coordinates:
                        out.append(str(g.start))
                        out.append(str(g.end))
                    if lengths:
                        out.append(str(g.size))
                    if coding_info:
                        out.append(str(g.coding))
                    if gene_symbols:
                        out.append("|".join(g.symbols))
                    f_out.write(sep.join(out) + "\n")

    def transcript_list(self, custom_path: str = "", output_file: str = "", lengths: bool = False, coordinates: bool = False, chromosomes: bool = False, coding_info: bool = False, skip_coding: bool = False, skip_non_coding: bool = False, sep: str = "\t", skip_pseudogenes: bool = False, skip_transposables: bool = False, gene_symbols: bool = False):
        if not custom_path:
            export_folder = Path(self._annot.path) / "lists"
        else:
            export_folder = Path(custom_path)
        export_folder.mkdir(parents=True, exist_ok=True)
        export_folder = str(export_folder) + "/"

        if not output_file:
            output_file = f"{export_folder}{self._annot.name}_transcripts.txt"
        else:
            output_file = f"{export_folder}{output_file}"

        if skip_coding and skip_non_coding:
            print("Warning: Both skip_coding and skip_non_coding are set to True. No transcripts will be listed.")
            return

        with open(output_file, "w", encoding="utf-8") as f_out:

            header = ["transcript_id"]
            if chromosomes or coordinates:
                header.append("scaffold")
            if coordinates:
                header.append("transcript_start")
                header.append("transcript_end")
            if lengths:
                header.append("transcript_length")
            if coding_info:
                header.append("coding")
            if gene_symbols:
                header.append("gene_symbol")
            f_out.write(sep.join(header) + "\n")

            for chrom, genes in self._annot.chrs.items():
                for g in genes.values():
                    if skip_pseudogenes and g.pseudogene:
                        continue
                    if skip_transposables and g.transposable:
                        continue

                    for t in g.transcripts.values():
                        if skip_coding and t.coding:
                            continue
                        if skip_non_coding and not t.coding:
                            continue

                        out = [t.id]
                        if chromosomes or coordinates:
                            out.append(chrom)
                        if coordinates:
                            out.append(str(t.start))
                            out.append(str(t.end))
                        if lengths:
                            out.append(str(t.size))
                        if coding_info:
                            out.append(str(t.coding))
                        if gene_symbols:
                            out.append("|".join(g.symbols))
                        f_out.write(sep.join(out) + "\n")





import typer
import os

from typing_extensions import Annotated

from ..annotation import Annotation
from ..genome import Genome

app = typer.Typer(add_completion=False, no_args_is_help=True)

@app.command()
def main(
    annotation_file: Annotated[str, typer.Argument(
        help="Path to the input annotation GFF/GTF file."
    )],
    genome_file: Annotated[str, typer.Argument(
        help="Path to the input genome FASTA file."
    )],
    annotation_name: Annotated[str, typer.Option(
        "-a", "--annotation-name", help="Annotation version, name or tag."
    )] = "{annotation-file}",
    genome_name: Annotated[str, typer.Option(
        "-g", "--genome-name", help="Genome assembly version, name or tag."
    )] = "{genome-file}",
    output_dir: Annotated[str, typer.Option(
        "-d", "--output-dir", help="Path to the output folder."
    )] = "./aegis_output/",
    quiet: Annotated[bool, typer.Option(
        "-q", "--quiet", help="Keeps terminal reporting to a minimum."
    )] = False,
):
    """
    Outputs a summary with descriptive stats.
    """

    if annotation_name == "{annotation-file}":
        annotation_name = os.path.splitext(os.path.basename(annotation_file))[0]

    os.makedirs(output_dir, exist_ok=True)

    annotation = Annotation(name=annotation_name, annot_file_path=annotation_file, quiet=quiet)
    genome = Genome(name=genome_name, genome_file_path=genome_file, quiet=quiet)

    annotation.stats.update(custom_path=output_dir, export=True, genome=genome, quiet=quiet)


if __name__ == "__main__":
    app()



"""
Tests for aegis.feature — the Feature base class.
"""

import pytest

from aegis.feature import Feature


# ============================================================
# Helpers
# ============================================================

def make_feature(**overrides):
    """Factory for Feature instances with sensible defaults."""
    defaults = dict(
        feature_id="feat001",
        ch="chr1",
        source="aegis",
        feature="gene",
        strand="+",
        start=100,
        end=500,
        score=".",
        phase=".",
        attributes={"Name":["TestFeature"], "Alias": ["TF1", "TF2"], "Symbol": ["TFS"]}
    )
    defaults.update(overrides)
    return Feature(**defaults)


# ============================================================
# __init__ and attribute parsing
# ============================================================

class TestFeatureInit:
    def test_basic_properties(self):
        f = make_feature()
        assert f.id == "feat001"
        assert f.original_id == "feat001"
        assert f.ch == "chr1"
        assert f.source == "aegis"
        assert f.feature == "gene"
        assert f.strand == "+"
        assert f.start == 100
        assert f.end == 500

    def test_size_calculation(self):
        f = make_feature(start=100, end=500)
        assert f.size == 401

    def test_names_parsed(self):
        f = make_feature()
        assert "TestFeature" in f.names

    def test_aliases_parsed(self):
        f = make_feature()
        assert "TF1" in f.aliases
        assert "TF2" in f.aliases

    def test_symbols_parsed(self):
        f = make_feature()
        assert "TFS" in f.symbols

    def test_id_number_extraction(self):
        f = make_feature(feature_id="gene123")
        assert f.id_number == 123
        assert f.original_id_number == 123

    def test_id_number_no_digits(self):
        f = make_feature(feature_id="gene_abc")
        assert f.id_number is None

    def test_dict_and_list_attributes(self):
        f = make_feature(feature_id="feat01", attributes={"Name": ["Test"]}, parents=["mRNA1"])
        # feature_id arg is used for self.id, dict is converted to attributes list
        assert f.id == "feat01"
        assert "mRNA1" in f.parents
        assert "Test" in f.names

    def test_misc_attributes_collected(self):
        f = make_feature(feature_id="feat1", attributes={"Dbxref": "GeneID:12345", "custom": "value"})
        assert any("Dbxref" in a for a in f.misc_attributes)
        assert any("custom" in a for a in f.misc_attributes)


# ============================================================
# Methods
# ============================================================

class TestFeatureMethods:
    def test_update_size(self):
        f = make_feature(start=100, end=500)
        f.end = 600
        f.update_size()
        assert f.size == 501

    def test_print_gff_format(self):
        f = make_feature(feature_id="feat001")
        gff_line = f.print_gff()
        assert gff_line.startswith("chr1\taegis\tgene\t100\t500")
        assert "ID=feat001" in gff_line
        assert gff_line.endswith("\n")

    def test_copy_is_independent(self):
        f = make_feature()
        f2 = f.copy()
        f2.id = "changed"
        assert f.id == "feat001"

    def test_str(self):
        f = make_feature()
        assert str(f) == "feat001"

    def test_calculate_gc_content(self):
        f = make_feature(start=1, end=10)
        f.seq = "GGCCAATTGG"  # 6 GC out of 10
        f.calculate_gc_content()
        assert f.gc_content == 0.6

    def test_calculate_gc_content_empty_seq(self):
        f = make_feature()
        f.seq = ""
        f.calculate_gc_content()
        assert f.gc_content == 0

    def test_clear_sequence(self):
        f = make_feature()
        f.seq = "ATGC"
        f.hard_seq = "ATGC"
        f.clear_sequence()
        assert f.seq == ""
        assert f.hard_seq == ""

    def test_clear_sequence_just_hard(self):
        f = make_feature()
        f.seq = "ATGC"
        f.hard_seq = "NNNN"
        f.clear_sequence(just_hard=True)
        assert f.seq == "ATGC"  # preserved
        assert f.hard_seq == ""


# ============================================================
# Comparison operators
# ============================================================

class TestFeatureComparisons:
    def test_equal_features(self):
        f1 = make_feature()
        f2 = make_feature()
        assert f1.identical(f2)

    def test_not_equal_different_id(self):
        f1 = make_feature(feature_id="a")
        f2 = make_feature(feature_id="b")
        assert not (f1.identical(f2))

    def test_lt_by_start(self):
        f1 = make_feature(start=100, end=500)
        f2 = make_feature(start=200, end=500)
        assert f1 < f2

    def test_lt_same_start_different_end(self):
        f1 = make_feature(start=100, end=400)
        f2 = make_feature(start=100, end=500)
        assert f1 < f2

    def test_le(self):
        f1 = make_feature(start=100, end=500)
        f2 = make_feature(start=100, end=500)
        assert f1 <= f2

    def test_equal_coordinates(self):
        f1 = make_feature(start=100, end=500, ch="chr1")
        f2 = make_feature(start=100, end=500, ch="chr1")
        assert f1.equal_coordinates(f2)

    def test_not_equal_coordinates_different_chr(self):
        f1 = make_feature(ch="chr1")
        f2 = make_feature(ch="chr2")
        assert not f1.equal_coordinates(f2)

    def test_equal_sequence_location(self):
        f1 = make_feature(start=100, end=500, ch="chr1", strand="+")
        f2 = make_feature(start=100, end=500, ch="chr1", strand="+")
        assert f1.equal_sequence(f2)

    def test_not_equal_sequence_different_strand(self):
        f1 = make_feature(strand="+")
        f2 = make_feature(strand="-")
        assert not f1.equal_sequence(f2)

    def test_longer(self):
        f1 = make_feature()
        f1.seq = "ATGCATGC"
        f2 = make_feature()
        f2.seq = "ATGC"
        assert f1.longer(f2) is True
        assert f2.longer(f1) is False



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..annotation import Annotation
    from ..genome import Genome

import pandas as pd

from pathlib import Path
from statistics import mean

from ..utils.plots import barplot, pie_chart
from ..subfeatures import Intron

class AnnotationStats:
    """
    Component for handling statistical methods and metric calculations for the Annotation class.
    Accessed via 'annotation_object.stats'.
    """
    data: dict

    def __init__(self, annotation:Annotation):
        self._annot = annotation
        self.data = {}

    def calculate_transcript_masking(self, hard_masked_genome:Genome):
        for genes in self._annot.chrs.values():
            for g in genes.values():
                g.generate_hard_sequence(hard_masked_genome)
                for t in g.transcripts.values():
                    t.generate_hard_sequence(hard_masked_genome)
                    t.calculate_masking()
                    t.clear_sequence(just_hard=True)
                    for c in t.CDSs.values():
                        c.generate_hard_sequence(hard_masked_genome)
                        c.calculate_masking()
                        c.clear_sequence(just_hard=True)
        self._annot.update()
    
    def calculate_gc_content(self):
        for genes in self._annot.chrs.values():
            for g in genes.values():
                g.calculate_gc_content()
                for t in g.transcripts.values():
                    t.calculate_gc_content()
                    for c in t.CDSs.values():
                        c.calculate_gc_content()
    def gene_count(self):
        gene_objects = 0
        unique_gene_ids_in_overlaps = set()
        for genes in self._annot.chrs.values():
            gene_objects += len(genes)
            for g in genes.values():
                for o in g.overlaps["self"]:
                    unique_gene_ids_in_overlaps.add(o.id)
        print(f"There are {gene_objects} gene objects and {len(self._annot.all_gene_ids)} genes in all gene ids and {len(unique_gene_ids_in_overlaps)} ids contained in self overlaps.")

    def update(self, custom_path:str="", export:bool=False, genome:Genome|None=None, max_x:int|None=None, quiet:bool=True):
        if not quiet:
            print(f"\nUpdating stats for {self._annot.id}")
        if not self._annot.generated_all_sequences or not self._annot.contains_protein_sequences:
            if genome != None:
                self._annot.generate_sequences(genome, quiet=quiet)
                self._annot.clear_sequences(keep_proteins=True)

        self.calculate_gc_content()

        if export:
            export_folder = Path(custom_path or self._annot.path) / "out_stats"
            export_folder.mkdir(parents=True, exist_ok=True)
            export_folder = str(export_folder) + "/"

        to_tally = ["coding_genes", "noncoding_genes", "CDSs_without_stop", "CDSs_with_stop"]

        self.data = {"mean_transcripts" : [], "mean_exons" : [], "mean_exon_size" : [], "mean_gene_size" : [], "mean_intron_size" : [], "mean_CDS_size" : [], "mean_UTR_size" : [], "mean_transcript_size" : [], "mean_five_prime_UTR_size" : [], "mean_three_prime_UTR_size" : []}

        if genome != None:
            self.data["mean_protein_size"] = []

        for key in to_tally:
            self.data[key] = []

        for ft, value in self._annot.features.items():
            self.data[ft] = value

        self.data["five_prime_UTRs"] = 0
        self.data["three_prime_UTRs"] = 0

        for genes in self._annot.chrs.values():
            for g in genes.values():
                self.data["mean_transcripts"].append(len(g.transcripts))
                self.data["mean_gene_size"].append(g.size)
                for t in g.transcripts.values():
                    if t.main:
                        if hasattr(t, "introns"):
                            for i in t.introns:
                                self.data["mean_intron_size"].append(i.size)

                        if t.coding:
                            for c in t.CDSs.values():
                                if c.main:
                                    self.data["mean_CDS_size"].append(c.size)
                                    if hasattr(c, "UTRs"):
                                        utr_5 = False
                                        utr_3 = False
                                        u_size = 0
                                        u5_size = 0
                                        u3_size = 0
                                        for u in c.UTRs:
                                            if u.prime == "3'":
                                                utr_3 = True
                                                u3_size += u.size
                                            else:
                                                utr_5 = True
                                                u5_size += u.size
                                            u_size += u.size
                                        if utr_5:
                                            self.data["five_prime_UTRs"] += 1
                                        if utr_3:
                                            self.data["three_prime_UTRs"] += 1
                                        self.data["mean_UTR_size"].append(u_size)
                                        self.data["mean_five_prime_UTR_size"].append(u5_size)
                                        self.data["mean_three_prime_UTR_size"].append(u3_size)
                                    if genome is not None and c.protein is not None:
                                        self.data["mean_protein_size"].append(c.protein.size)
                            self.data["coding_genes"].append(g.id)
                        else:
                            self.data["noncoding_genes"].append(g.id)

                        for e in t.exons:
                            self.data["mean_exon_size"].append(e.size)

                        self.data["mean_exons"].append(len(t.exons))
                        
                        self.data["mean_transcript_size"].append(t.size)

        # anything with mean will be also plotted as distribution plots:
        if export:
            for key in self.data:
                if "mean" in key:
                    tag = key.split("mean_")[1]
                    if tag[-1] != "s":
                        tag += "s"
                    barplot(self.data[key], export_folder, f"{self._annot.id}{self._annot.feature_suffix}_{tag}", f"Distribution of {self._annot.id} {tag}", max_x)   

        if genome != None:
            self.data["CDSs_without_stop"] = []
            self.data["CDSs_with_stop"] = []
            self.data["intron_composition"] = set()
            intron_stats = {}
            for b in Intron.canonical_seqs:
                intron_stats[f"intron-exon boundary: {b}"] = 0
            intron_stats["other_intron_seqs"] = 0
            for genes in self._annot.chrs.values():
                for g in genes.values():                    
                    for t in g.transcripts.values():
                        if t.main:
                            if t.coding:
                                for c in t.CDSs.values():
                                    if c.main:
                                        if c.protein != None:
                                            if not c.protein.end_stop:
                                                self.data["CDSs_without_stop"].append(g.id)
                                            else:
                                                self.data["CDSs_with_stop"].append(g.id)
                                            
                            for i in t.introns:
                                self.data["intron_composition"].add(i.boundary)
                                if i.canonical:
                                    intron_stats[f"intron-exon boundary: {i.boundary}"] += 1
                                else:
                                    intron_stats["other_intron_seqs"] += 1
            self.data["intron_composition"] = list(self.data["intron_composition"])

            if export:
                labels = list(intron_stats.keys())
                mod_labels = []
                for l in labels:
                    if "other" in l:
                        mod_labels.append("other")
                    else:
                        mod_labels.append(l.split(": ")[1])

                values = list(intron_stats.values())
                pie_chart(mod_labels, values, export_folder, f"{self._annot.id}{self._annot.feature_suffix}_intron_composition", f"Intron composition of {self._annot.id} annotation")
            self.data.update(intron_stats)

        for key in self.data:
            if "mean" in key:
                if len(self.data[key]) > 0:
                    self.data[key] = mean(self.data[key])
                else:
                    self.data[key] = 0
            #tallying
            elif key in to_tally:
                self.data[key] = len(self.data[key])

        if export:
            if genome != None:
                out_file = f"{self._annot.id}{self._annot.feature_suffix}_full_stats.csv"
            else:
                out_file = f"{self._annot.id}{self._annot.feature_suffix}_basic_stats.csv"

            warning_df = pd.DataFrame({k: pd.Series(sorted(v)) for k, v in self._annot.warnings.items()}).fillna('')
            error_df = pd.DataFrame({k: pd.Series(sorted(v)) for k, v in self._annot.errors.items()}).fillna('')

            warning_df.to_csv(f"{export_folder}{self._annot.id}{self._annot.feature_suffix}_warnings.csv", sep="\t", index=False)
            error_df.to_csv(f"{export_folder}{self._annot.id}{self._annot.feature_suffix}_errors.csv", sep="\t", index=False)

            f_out = open(f"{export_folder}{out_file}", "w", encoding="utf-8")
            f_out.write("")
            f_out.close()

            f_out = open(f"{export_folder}{out_file}", "a")
            x = -1
            for key, value in self.data.items():
                x += 1
                value_temp = value
                if isinstance(value, list):
                    if len(value) > 200:
                        value_temp = value[:199]
                        value_temp.append("...")
                if x == 0:
                    f_out.write(f"{key}\t{value_temp}")
                else:
                    f_out.write(f"\n{key}\t{value_temp}")
            f_out.close()
        if not quiet:
            print(f"\nUpdated stats for {self._annot.id}")



import typer
import os
import pandas as pd

from typing_extensions import Annotated

from ..annotation import Annotation
from ..genome import Genome


app = typer.Typer(add_completion=False, no_args_is_help=True)

@app.command()
def main(
    annotation_file: Annotated[str, typer.Argument(
        help="Path to the input annotation GFF/GTF file."
    )],
    genome_file: Annotated[str, typer.Argument(
        help="Path to the input genome FASTA file."
    )],
    genelist: Annotated[str, typer.Argument(
        help="Input tsv file with list of 'gene-id' entries. Excel '.xlsx' files are also permitted."
    )],
    motif: Annotated[str, typer.Argument(
        help="Regular expression pattern in python describing motif."
    )],
    motif_length: Annotated[int, typer.Argument(
        help="Actual length of motif."
    )],
    header: Annotated[bool, typer.Option(
        "-H", "--header", help="Use this flag to indicate the presence of a header in input genelist file."
    )] = False,
    promoter_size: Annotated[int, typer.Option(
        "-ps", "--promoter-size", help=f"Only applies if promoter included in '-f'. Promoter size in bp upstream of TSS or ATG depending on '-p'."
    )] = 2000,
    promoter_type: Annotated[str, typer.Option(
        "-p", "--promoter-type", help=f"Only applies if promoter included in '-f'. Defines the reference point for the promoter regions of '-ps' size. 'standard': Generated upstream of the transcript's start site (TSS); 'upstream_ATG': Generated upstream of the main CDS's start codon (ATG). If no CDS, falls back to standard; 'standard_plus_up_to_ATG': Generated upstream of the transcript's start site (TSS) and any gene sequence up to the start codon (ATG) is also added. If no CDS, falls back to standard."
    )] = "standard",
    annotation_name: Annotated[str, typer.Option(
        "-a", "--annotation-name", help="Annotation version, name or tag."
    )] = "{annotation-file}",
    genome_name: Annotated[str, typer.Option(
        "-g", "--genome-name", help="Genome assembly version, name or tag."
    )] = "{genome-file}",
    query_tag: Annotated[str, typer.Option(
        "--genelist-tag", help="Query gene list tag/name to improve output description."
    )] = "query_genes",
    motif_tag: Annotated[str, typer.Option(
        "--motif-tag", help="Motif tag/name to improve output description, e.g. '{TF}_{motif_name}'."
    )] = "query_motif",
    output_dir: Annotated[str, typer.Option(
        "-d", "--output-dir", help="Path to the output directory."
    )] = "./aegis_output/",
    quiet: Annotated[bool, typer.Option(
        "-q", "--quiet", help="Keeps terminal reporting to a minimum."
    )] = False,

):
    """
    Scans a set of “query” genes to locate all occurrences of a specified DNA motif within their upstream promoter regions.
    """

    if annotation_name == "{annotation-file}":
        annotation_name = os.path.splitext(os.path.basename(annotation_file))[0]

    os.makedirs(output_dir, exist_ok=True)

    if header:
        if genelist.endswith(".xlsx"):
            df = pd.read_excel(genelist, skiprows=1, dtype=str)
        else:
            df = pd.read_csv(genelist, skiprows=1, dtype=str)
    else:
        if genelist.endswith(".xlsx"):
            df = pd.read_excel(genelist, dtype=str)
        else:
            df = pd.read_csv(genelist, dtype=str)

    df = df.fillna("")
    genes = df.iloc[:, 0].tolist()
    genes = [gene for gene in genes if gene != ""]

    annotation = Annotation(name=annotation_name, annot_file_path=annotation_file, quiet=quiet)
    genome = Genome(name=genome_name, genome_file_path=genome_file, quiet=quiet)

    annotation.generate_promoters(genome=genome, promoter_size=promoter_size, promoter_type=promoter_type, generate_sequence=True)

    annotation.motifs.find_and_plot(query_genes=genes, motif=motif, motif_length=motif_length, glistname=query_tag, tf_motif_tag=motif_tag, custom_path=output_dir, quiet=quiet)
    
if __name__ == "__main__":
    app()
